/***********************************************************************
 HVProbe.c - Multi-platform library for communication
 with CPS USB-compatible high-voltage probes.

 Jim Lamberson
 CPS

 OS-independent API core

 Copyright 2015, All Rights Reserved.

 This file is part of the HVProbe Application Program Interface (API).
 The API is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 The API is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with the API. If not, see <http://www.gnu.org/licenses/>.

************************************************************************/

#define BUILD_API           // export DLL functions

#include <string.h>
#include <math.h>
#include <memory.h>

#include "HVProbe.h"
#include "hidapi.h"

#ifdef __cplusplus
extern "C" {
#endif

// API version numbers =======================================

#define VERSION_MAJOR   2
#define VERSION_MINOR   0
#define VERSION_BUILD   0

// Version history:
//  ---     Initial release.
//  1.1.0   Added 2 functions: HVP_ReadApiVersion(), HVP_ErrorString().
//  2.0.0   Converted HID interface to hidapi to simplify porting to linux and mac

// Constants ------------------------------------------------------

#define TRUE                        1
#define FALSE                       0

#define STD_CAL_T                   38.0        // Calibration reference temperature

// Size in bytes of USB buffers
#define USB_BUFSIZE                 64          // USB buffer size -- excludes extra ReportNumber byte used by some (e.g., Microsoft) USB implementations (first byte)

// USB device command opcodes
#define DEVCMD_GET_PROPERTIES       0x20        // Read device properties
#define DEVCMD_GET_PARAM            0x21        // Read one device parameter
#define DEVCMD_GET_FIRMWARE_INFO    0x22        // Read firmware info string
#define DEVCMD_IDENTIFY_DEVICE      0x80        // Read device identity
#define DEVCMD_GET_STATUS           0x81        // Read alarm status (probe ground lead is disconnected)
#define DEVCMD_GET_VALUES           0x82        // Read all meter values
#define DEVCMD_READ_EEPROM          0x83        // Read all EEPROM data
#define DEVCMD_WRITE_EEPROM         0x84        // Write all EEPROM data
#define DEVCMD_SET_STATUS           0x85
#define DEVCMD_SET_VALUES           0x86

// Byte offsets of data returned by DEVCMD_GET_VALUES
#define T_BUF_OFFSET                1
#define DC_BUF_OFFSET               5
#define AC_BUF_OFFSET               9
#define MAX_BUF_OFFSET              13
#define MIN_BUF_OFFSET              17

// Types ---------------------------------------------

typedef unsigned char u8;

enum device_types { v_meter, a_meter, va_meter, v_control, a_control, va_control, invalid_device_type };
enum param_types  { none, voltage, current, temperature, invalid_param_type };
enum param_direction  { input, output, inout, invalid_param_direction };
enum param_units  { nV, uV, mV, V, kV, MV, pA, nA, uA, mA, A, kA, C, K, invalid_param_unit };

typedef enum PARAM_ID {         // USB DEVICE PARAMETER IDENTIFIERS -----------
    PARAM_T,                    //  Temperature
    PARAM_DC,                   //  DC voltage
    PARAM_AC,                   //  AC voltage
    NUM_PARAMS
} PARAM_ID;

typedef struct DEV_PROPERTIES {
    int model;              // model number
    u8  type;               // type, one of device_types
    u8  version;            // device version
    u8  nparams;            // number of params
} DEV_PROPERTIES;

typedef struct EEPROM_IMAGE {   // EEPROM CONTENTS IN DEVICE BYTE ORDER ---------------
                                // Cal values used to convert adc to engineering units:
    int dc_scale;               //  Scalar
    int dc_offset;              //  Offset
    int dc_scale_t;             //  Delta-temperature scalar
    int dc_offset_t;            //  Delta-temperature offset
                                // Cal values used to convert adc to engineering units:
    int ac_scale;               //  Scalar
    int ac_offset;              //  Offset
    int ac_scale_t;             //  Delta-temperature scalar
    int ac_offset_t;            //  Delta-temperature offset
                                // Cal values for linearization corrections:
    int dc_scale2;              //  DC scalar
    int ac_scale2;              //  AC scalar
                                // Unused:
    int dc_scale3;
    int ac_scale3;
    int reserve1;
    int reserve2;
                                // Other:
    int serial;                 //  Probe serial number
    u8  checksum;               //  EEPROM checksum
} EEPROM_IMAGE;

typedef struct CAL_DATA {       // CALIBRATION DATA IN HOST BYTE ORDER ---------------
                                // Cal values used to convert adc to engineering units:
    float dc_scale;             //  Scalar
    float dc_offset;            //  Offset
    float dc_scale_t;           //  Delta-temperature scalar
    float dc_offset_t;          //  Delta-temperature offset
                                // Cal values used to convert adc to engineering units:
    float ac_scale;             //  Scalar
    float ac_offset;            //  Offset
    float ac_scale_t;           //  Delta-temperature scalar
    float ac_offset_t;          //  Delta-temperature offset
                                // Cal values for linearization corrections:
    float dc_scale2;            //  DC scalar
    float ac_scale2;            //  AC scalar
} CAL_DATA;

#define EEPROM_NBYTES   61

typedef struct PARAM_INFO {     // USB DEVICE PARAMETER INFO (FOR ONE PARAMETER) --------------
                                // Values reported by device:
    u8      type;               //  type, one of param_types
    u8      dir;                //  directions, one of param_direction
    u8      unit;               //  user's unit, one of param_units
    u8      bits;               //  size in bits of right-justified native adc value
    u8      is_signed;          //  true if adc value is signed
    int     max_binary;         //  binary value corresponding to maximum engineering units value
    int     zero_binary;        //  binary value corresponding to zero engineering units value
    int     min_binary;         //  binary value corresponding to minimum engineering units value
    int     over_binary;        //  overflow indicated if adc value equals this
    int     under_binary;       //  underflow indicated if adc value equals this
    double  max_user;           //  maximum value in engineering units
    double  zero_user;          //  zero value in engineering units
    double  min_user;           //  minimum value in engineering units
                                // Computed values:
    double  engr_units_scalar;
    double  engr_units_offset;
    int     engr_units_nshift;
} PARAM_INFO;

typedef struct METER {              // INDIVIDUAL VOLTMETER ------------------
    int         usb_buf_offset;     //  Byte offset of received data in USB buffer
    PARAM_INFO  *param;             //  Pointer to associated device parameters
    int         is_overrange;       //  Boolean: adc saturation indicator
    double      uncorr;             //  Uncorrected meter sample, in engineering units
    double      corr;               //  Corrected meter sample
} METER;

typedef struct PROBE {                              // PROBE INSTANCE INFO ---------
    int             is_open;                        //  Boolean: probe is open
    int             ProbeID;                        //  USB enumerated probe number
    hid_device      *hProbe;                        //  Probe's HID API info
    DEV_PROPERTIES  device_properties;              //  USB device properties
    PARAM_INFO      param_info[NUM_PARAMS];         //  USB device parameters
    double          ac_lastv;                       //  Last AC sample; used to suppress AC zero offset
    double          delta_t;                        //  Difference between current temperature and cal ref temp in degrees C
    EEPROM_IMAGE    eebuf;                          //  EEPROM image
    CAL_DATA        caldata;                        //  Calibration data
    PROBE_INFO      probe_info;                     //  Public probe attributes
    METER           meter[NUM_METERS];              //  The voltmeters
} PROBE;


// Storage ---------------------------------------------------------------

static int          is_hid_open = 0;
static PROBE        probe[MAX_NUM_PROBES];                  // Probe attributes and data


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////  PRIVATE FUNCTIONS  /////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// Convert a 4-byte value from device to host endian.

static const int one = 1;
#define IS_LITTLE_ENDIAN_CPU ((u8)one)  // endian test

static int DevToHostEndian(void *buf)
{
    if (IS_LITTLE_ENDIAN_CPU)
        return ((u8*)buf)[0] | (((u8*)buf)[1] << 8) | (((u8*)buf)[2] << 16) | (((u8*)buf)[3] << 24);    // big->little endian
    else
        return *(int *)buf;     // host/device have same endianness
}

///////////////////////////////////////////////////////////
// Convert 4-byte device value to float.

static float DevQuadletToFloat(void *quadlet)
{
    union {
        int     i;
        float   f;
    } u;

    u.i = DevToHostEndian(quadlet);         // convert to host endian int
    return u.f;                             // cast as float and return
}

///////////////////////////////////////////////////////////////////////
// Get the probe instance of an open device.

static int GetLiveProbe(int ProbeID, PROBE **pprobe)
{
    if (ProbeID > (MAX_NUM_PROBES - 1))     // Abort with error if probe ID is invalid.
        return HVP_ERR_PROBE_ID;

    if (!probe[ProbeID].is_open)            // Abort with error if device is closed.
        return HVP_ERR_PROBE_CLOSED;

    *pprobe = &probe[ProbeID];              // Return probe instance.
    return HVP_ERR_OK;
}

////////////////////////////////////////////////////////////////
// Execute USB transaction with probe.

static int UsbTransact(PROBE *p, u8 *outbuf, u8 *inbuf)
{
    // In all hid api implementations, hid_write() always expects an extra 0 byte to precede the 64 data bytes in outbuf.
    unsigned char txbuf[65];
    txbuf[0] = 0;

    memcpy(txbuf + 1, outbuf, 64);

    if (hid_write(p->hProbe, txbuf, 65) != 65)      // Send command to probe.
        return HVP_ERR_WRITE;                       // Abort if USB write failed.

    if (inbuf == NULL)                              // If no reply is expected from probe
        return HVP_ERR_OK;                          //   the transaction is finished.

    // Receive reply from probe. Note: hid_read() drops the extra 0 byte preceding the 64 data bytes in some USB implementations (e.g., Windows).
    return (hid_read(p->hProbe, inbuf, 64) == 64) ? HVP_ERR_OK : HVP_ERR_READ;
}

////////////////////////////////////////////////////////////
// Compute EEPROM checksum over specified range.

static u8 CalcChecksum(u8 *eebuf, int nbytes)
{
    u8 checksum = 0;
    for (; nbytes > 0; nbytes--)
        checksum += *eebuf++;
    return checksum;
}

//////////////////////////////////////////////////////////////////
// Read data from device EEPROM into buf in device byte order.

static int ReadEeprom(PROBE *p, u8 *buf)
{
    u8 inbuf[USB_BUFSIZE];
    u8 cmdbuf[USB_BUFSIZE] = {DEVCMD_READ_EEPROM};

    int err = UsbTransact(p, cmdbuf, inbuf);            // Send command to probe
    if (err == HVP_ERR_OK)
        memcpy(buf, &inbuf[1], EEPROM_NBYTES);          // Copy received EEPROM data to buf
    return err;
}

/////////////////////////////////////////////////////////////////
// Write data (must be in device byte order) to probe EEPROM.
// Note: checksum is automatically computed and inserted.

static int WriteEeprom(PROBE *p, u8 *buf)
{
    u8 outbuf[USB_BUFSIZE] = {DEVCMD_WRITE_EEPROM};

    memcpy(&outbuf[2], buf, EEPROM_NBYTES - 1);                         // Copy EEPROM data to usb buffer
    outbuf[EEPROM_NBYTES + 1] = -CalcChecksum(buf, EEPROM_NBYTES - 1);  // Compute and insert EEPROM checksum

    return UsbTransact(p, outbuf, NULL);        // Send command and data to probe; no reply is expected
}

//////////////////////////////////////////////////////////////
// Correct AC zero offset and filter to reduce AC noise.

static double correctacoffset(PROBE *p, double v)
{
    double dv;
    double maxv = p->param_info[PARAM_AC].max_user;
    double relv = v / maxv;
    double reloffset = -p->caldata.ac_offset / maxv;
    double flt = 0.01 - fabs(relv);
    flt = (flt > 0) ? (flt * 1000 + 1) : 1;

    dv = reloffset / ((relv - reloffset) * 500 + 1);

    v -= dv * maxv;

    dv = (v - p->ac_lastv) / flt;

    p->ac_lastv += dv;

    return p->ac_lastv;
}

//////////////////////////////////////////////////
// Calibration-correct an AC meter sample.

static void CorrectAC(PROBE *p)
{
    METER *m = &p->meter[METER_AC];

    if (m->is_overrange)            // If adc saturated
        m->corr = m->uncorr;        //   don't apply cal corrections
    else                            // Else
    {
        m->corr = ((m->uncorr * (p->caldata.ac_scale + 1) + ((m->uncorr > 0) ? 1.0 : -1.0) * m->uncorr * m->uncorr * p->caldata.ac_scale2))    // scale and linearize
            * (1 + p->caldata.ac_scale_t * p->delta_t);          // correct for deviation from cal temperature

        m->corr = correctacoffset(p, m->corr);        // filter AC meter and attempt to suppress AC zero offset
    }
}

//////////////////////////////////////////////////////////////
// Calibration-correct a DC or peak-detect meter sample.

static void CorrectDC(PROBE *p, METER_ID meter_id)
{
    METER *m = &p->meter[meter_id];

    if (m->is_overrange)            // If adc saturated
        m->corr = m->uncorr;        //   don't apply cal corrections
    else                            // Else
    {
        m->corr = ((m->uncorr * (p->caldata.dc_scale + 1) + ((m->uncorr > 0) ? 1.0 : -1.0) * m->uncorr * m->uncorr * p->caldata.dc_scale2))    // scale and linearize
            * (1 + p->caldata.dc_scale_t * p->delta_t)    // correct for deviation from cal temperature
            + p->caldata.dc_offset;                       // correct fixed calibration offset
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////
// Calibration-correct a temperature meter sample. This also calculates and stores delta_t, which is
// the difference between probe and reference temperature (used to temperature-correct other meters).

static void CorrectT(PROBE *p)
{
    METER *m = &p->meter[METER_T];

    if ((m->uncorr < m->param->min_user) || (m->uncorr > m->param->max_user))   // If measured temperature is not a reasonable value
        m->corr = STD_CAL_T;                                                    //   use calibration reference temperature as corrected temp.
    else                                                                        // Else
        m->corr = m->uncorr;                                                    //   set corrected temp to uncorrected temp, which is expressed in degrees C

    p->delta_t = m->corr - STD_CAL_T;      // Compute and store delta-t from cal reference temperature.
}

/////////////////////////////////////////////////////////////
// Fetch calibration-corrected sample data into corr.

static void CorrectMeters(PROBE *p)
{
    CorrectT(p);               // Correct temperature. THIS MUST BE THE FIRST CORRECTION, because it computes delta_t, which is used to correct all other meters.
    CorrectDC(p, METER_DC );   // Correct DC voltage
    CorrectAC(p);              // Correct AC RMS voltage
    CorrectDC(p, METER_MIN);   // Correct DC voltage peak minimum
    CorrectDC(p, METER_MAX);   // Correct DC voltage peak maximum
}

///////////////////////////////////////////////////////////////////////////////////////
// Fetch data from all probe meters and convert to uncorrected engineering units.

static int GetRawMeters(PROBE *p)
{
    int i;
    u8 inbuf[USB_BUFSIZE];
    u8 outbuf[USB_BUFSIZE] = {DEVCMD_GET_VALUES};

    int err = UsbTransact(p, outbuf, inbuf);                        // Fetch data from all probe meters via USB.
    if (err == HVP_ERR_OK)
    {
        METER *m = p->meter;
        for (i = 0; i < NUM_METERS; i++, m++)                       // For each meter:
        {
            PARAM_INFO *parm = m->param;

            int bin = DevToHostEndian(&inbuf[m->usb_buf_offset]);       // Convert adc value from usb to host endian.

            if (bin == parm->over_binary)                               // If adc is reporting overrange
            {
                m->is_overrange = TRUE;                                 //   indicate adc saturation
                m->uncorr = parm->max_user;                             //   return max legal value
            }
            else if (bin == parm->under_binary)                         // ElseIf adc is reporting underrange
            {
                m->is_overrange = TRUE;                                 //   indicate adc saturation
                m->uncorr = parm->min_user;                             //   return min legal value
            }
            else                                                        // Else
            {
                m->is_overrange = FALSE;                                //   indicate adc in range

                // Convert right-justified n-bit native adc value to sign-extended integer.
                bin <<= parm->engr_units_nshift;
                bin = parm->is_signed ? (bin >> parm->engr_units_nshift) : (int)((unsigned int)bin >> parm->engr_units_nshift);

                // Convert sign-extended adc value to engineering units.
                m->uncorr = bin * parm->engr_units_scalar + parm->engr_units_offset;
            }
        }
    }
    return err;
}

/////////////////////////////////////////////////////////////////////////////////////
// Read device properties/parameters and firmware info into private API storage.

static int GetProbeInfo(PROBE *p)
{
    int i;
    u8 inbuf[USB_BUFSIZE];

    // USB device commands
    u8 cmd_getprops[USB_BUFSIZE]  = {DEVCMD_GET_PROPERTIES};     // Get device properties
    u8 cmd_getparam[USB_BUFSIZE]  = {DEVCMD_GET_PARAM};          // Get device parameters (note: parameter number is specified by cmd_getparam[1])
    u8 cmd_getfwinfo[USB_BUFSIZE] = {DEVCMD_GET_FIRMWARE_INFO};  // Get firmware info string

    // Get device properties -----------------------------

    int err = UsbTransact(p, cmd_getprops, inbuf);                                              // Read device properties.
    if (err != HVP_ERR_OK)
        return err;

    memcpy(&p->device_properties, &inbuf[1], sizeof(DEV_PROPERTIES));                           // Copy properties to probe instance.

    if ((p->device_properties.type != v_meter) || (p->device_properties.nparams > NUM_PARAMS))  // Abort if device is not a voltmeter (apparently it's not an HV probe).
        return HVP_ERR_DEVICE_PARAM;

    // Get device parameters ------------------------------

    for (i = 0; i < p->device_properties.nparams; i++)          // For each device parameter ...
    {
        PARAM_INFO *parm = &p->param_info[i];

        cmd_getparam[1] = i;  // param number
        UsbTransact(p, cmd_getparam, inbuf);                        // Get parameter from device
        if (err != HVP_ERR_OK)
            return err;

        parm->type           = inbuf[1];                            // Copy parameter to probe instance
        parm->dir            = inbuf[2];
        parm->unit           = inbuf[3];
        parm->bits           = inbuf[4];
        parm->is_signed      = inbuf[5];
        parm->max_binary     = DevToHostEndian(&inbuf[6]);
        parm->zero_binary    = DevToHostEndian(&inbuf[10]);
        parm->min_binary     = DevToHostEndian(&inbuf[14]);
        parm->over_binary    = DevToHostEndian(&inbuf[18]);
        parm->under_binary   = DevToHostEndian(&inbuf[22]);
        parm->max_user       = (double)DevToHostEndian(&inbuf[26]);
        parm->zero_user      = (double)DevToHostEndian(&inbuf[30]);
        parm->min_user       = (double)DevToHostEndian(&inbuf[34]);

        // Abort if invalid parameter detected
        if ((parm->type >= invalid_param_type) || (parm->dir >= invalid_param_direction) || (parm->bits < 8) || (parm->bits > 32))
            return HVP_ERR_DEVICE_PARAM;

        // Compute binary shift count used to convert native n-bit adc value to integer.
        parm->engr_units_nshift = (sizeof(int) * 8) - parm->bits;

        // Compute scalar and offset used to convert integer adc value to engineering units.
        parm->engr_units_scalar = (parm->max_user - parm->min_user) / (parm->max_binary - parm->min_binary);
        parm->engr_units_offset = parm->max_user - parm->max_binary * parm->engr_units_scalar;
    }

    // Get firmware info ----------------------------

    UsbTransact(p, cmd_getfwinfo, inbuf);
    if (err == HVP_ERR_OK)
    {
        inbuf[USB_BUFSIZE - 1] = 0; // in case end-of-string missing
        strcpy(p->probe_info.firmware_info, (const char *)&inbuf[1]);
    }

    return err;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////  PUBLIC API FUNCTIONS  /////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////
// Enumerate and enable read/write access to all probes.
// Returns count of detected probes.

APIFUNC int CCONV HVP_OpenSystem(void)
{
    int nprobes = 0;
    struct hid_device_info *dev, *devlist;

    if (!hid_init())                                // Open the hid api. If successful ...
    {
        is_hid_open = 1;                                // Mark hid api Open.
        devlist = hid_enumerate(0x1943, 0x9001);        // Enumerate all probes in linked list (vendor_id 0x1943 = Sensoray, device_id 0x9001 = HV probe).

        for (nprobes = 0, dev = devlist; (nprobes < MAX_NUM_PROBES) && (dev != NULL); nprobes++, dev = dev->next)
            probe[nprobes].hProbe = hid_open_path(dev->path); // Open all enumerated probes.

        hid_free_enumeration(devlist);                  // Free probe enumeration linked list.
    }

    return nprobes;                                 // Return number of detected probes
}

///////////////////////////////////////////////////////////////////////////////////
// Close all probes and free API resources allocated by HVP_OpenSystem().

APIFUNC int CCONV HVP_CloseSystem(void)
{
    int i;
    if (is_hid_open)
    {
        for (i = 0; i < MAX_NUM_PROBES; i++)
        {
            probe[i].is_open = FALSE;

            if (probe[i].hProbe != NULL) {
                hid_close(probe[i].hProbe);
                probe[i].hProbe = NULL;
            }
        }
        hid_exit();
        is_hid_open = 0;
    }
    return HVP_ERR_OK;
}

////////////////////////////////////////////////////////////////////////////
// Open a probe. This initializes a probe instance and marks it open.

APIFUNC int CCONV HVP_OpenProbe(int ProbeID)
{
    int err;
    PROBE *p;

    if (ProbeID > (MAX_NUM_PROBES - 1))                     // Abort if invalid probe ID.
        return HVP_ERR_PROBE_ID;

    p = &probe[ProbeID];                                    // Point to probe instance.

    p->ProbeID = ProbeID;                                   // Save probe ID for reverse lookup.

    p->ac_lastv = 0;                                        // Prime AC RMS filter.

    err = GetProbeInfo(p);                                  // Read probe properties, parameters and firmware info from device.
    if (err != HVP_ERR_OK)
        return err;

   // Initialize calibration data ----------------------------

    err = ReadEeprom(p, (u8 *)&p->eebuf);                   // Read EEPROM contents from device (in device byte order).
    if (err != HVP_ERR_OK)
        return err;

    if (CalcChecksum((u8 *)&p->eebuf, EEPROM_NBYTES) != 0)  // If EEPROM checksum is invalid
    {
        memset(&p->eebuf, 0, EEPROM_NBYTES);                //   fill EEPROM image with default data (all zeros)
        err = HVP_ERR_CHECKSUM;                             //   indicate checksum error
    }

    // Extract calibration data from EEPROM image
    p->caldata.ac_offset    = DevQuadletToFloat(&p->eebuf.ac_offset);       // For each cal value in eeprom image,
    p->caldata.ac_offset_t  = DevQuadletToFloat(&p->eebuf.ac_offset_t);     //   convert to host endian and cast as float.
    p->caldata.ac_scale     = DevQuadletToFloat(&p->eebuf.ac_scale);
    p->caldata.ac_scale2    = DevQuadletToFloat(&p->eebuf.ac_scale2);
    p->caldata.ac_scale_t   = DevQuadletToFloat(&p->eebuf.ac_scale_t);
    p->caldata.dc_offset    = DevQuadletToFloat(&p->eebuf.dc_offset);
    p->caldata.dc_offset_t  = DevQuadletToFloat(&p->eebuf.dc_offset_t);
    p->caldata.dc_scale     = DevQuadletToFloat(&p->eebuf.dc_scale);
    p->caldata.dc_scale2    = DevQuadletToFloat(&p->eebuf.dc_scale2);
    p->caldata.dc_scale_t   = DevQuadletToFloat(&p->eebuf.dc_scale_t);

    // Populate remainder of public probe attributes ----------------

    // Get basic device info
    p->probe_info.model     = p->device_properties.model;
    p->probe_info.version   = p->device_properties.version;
    p->probe_info.serialnum = DevToHostEndian(&p->eebuf.serial);

    // Set scales of analog meters
    p->probe_info.vscale_dc = p->param_info[PARAM_DC].max_user;
    p->probe_info.vscale_ac = p->param_info[PARAM_AC].max_user;

    // Populate meter constants --------------------

    // Parameter info used by meter
    p->meter[METER_DC ].param = &p->param_info[PARAM_DC];
    p->meter[METER_AC ].param = &p->param_info[PARAM_AC];
    p->meter[METER_MIN].param = &p->param_info[PARAM_DC];
    p->meter[METER_MAX].param = &p->param_info[PARAM_DC];
    p->meter[METER_T  ].param = &p->param_info[PARAM_T ];

    // USB buffer offset of meter data
    p->meter[METER_DC ].usb_buf_offset = DC_BUF_OFFSET;
    p->meter[METER_AC ].usb_buf_offset = AC_BUF_OFFSET;
    p->meter[METER_MIN].usb_buf_offset = MAX_BUF_OFFSET;
    p->meter[METER_MAX].usb_buf_offset = MIN_BUF_OFFSET;
    p->meter[METER_T  ].usb_buf_offset = T_BUF_OFFSET;

    p->is_open = TRUE;  // indicate probe is open

    return err;
}

///////////////////////////////////////////////////////////
// Close a previously opened probe.

APIFUNC int CCONV HVP_CloseProbe(int ProbeID)
{
    PROBE *p;
    int err = GetLiveProbe(ProbeID, &p);    // Select probe instance
    if (err == HVP_ERR_OK)
        p->is_open = FALSE;     // mark probe closed
    return err;
}

////////////////////////////////////////////////
// Read probe EEPROM.

APIFUNC int CCONV HVP_ReadEeprom(int ProbeID, u8 *buf)
{
    PROBE *p;
    int err = GetLiveProbe(ProbeID, &p);   // point to probe struct
    if (err == HVP_ERR_OK)
        err = ReadEeprom(p, buf);
    return err;
}

///////////////////////////////////////////////////////////
// Write probe EEPROM.
// Note: checksum is automatically computed and inserted.

APIFUNC int CCONV HVP_WriteEeprom(int ProbeID, u8 *buf)
{
    PROBE *p;
    int err = GetLiveProbe(ProbeID, &p);
    if (err == HVP_ERR_OK)
        err = WriteEeprom(p, buf);
    return err;
}

//////////////////////////////////////////////////////////////
// Read probe status bits.

APIFUNC int CCONV HVP_GetStatus(int ProbeID, int *status)
{
    PROBE *p;
    u8 inbuf[USB_BUFSIZE];
    u8 outbuf[USB_BUFSIZE] = {DEVCMD_GET_STATUS};

    int err = GetLiveProbe(ProbeID, &p);    // Select probe instance
    if (err != HVP_ERR_OK)
        return err;

    UsbTransact(p, outbuf, inbuf);          // Read status byte from device
    if (err == HVP_ERR_OK)
        *status = (int)inbuf[1];

    return err;
}

//////////////////////////////////////////////////////////////////////////////
// Read uncorrected sample data (expressed in engineering units).

APIFUNC int CCONV HVP_GetRawMeters(int ProbeID, double buf[NUM_METERS])
{
    PROBE *p;
    int i;
    int err = GetLiveProbe(ProbeID, &p);        // Select probe instance
    if (err == HVP_ERR_OK)
    {
        err = GetRawMeters(p);                  // Read uncorrected meter data from probe
        for (i = 0; i < NUM_METERS; i++)        // Copy meter data to user buffer
            buf[i] = p->meter[i].uncorr;
    }
    return err;
}

/////////////////////////////////////////////////////////////
// Read calibration-corrected sample data.

APIFUNC int CCONV HVP_GetMeters(int ProbeID, double buf[NUM_METERS])
{
    PROBE *p;
    int i;
    int err = GetLiveProbe(ProbeID, &p);            // Select probe instance
    if (err != HVP_ERR_OK)
        return err;

    err = GetRawMeters(p);                      // Read uncorrected meter data from probe
    if (err == HVP_ERR_OK)
    {
        CorrectMeters(p);
        for (i = 0; i < NUM_METERS; i++)        // Copy meter data to user buffer
            buf[i] = p->meter[i].corr;
    }

    return err;
}

////////////////////////////////////////////////////////////
// Read public probe attributes from probe instance.

APIFUNC int CCONV HVP_GetAttributes(int ProbeID, PROBE_INFO *info)
{
    PROBE *p;
    int err = GetLiveProbe(ProbeID, &p);    // Select probe instance
    if (err == HVP_ERR_OK)
        *info = probe[ProbeID].probe_info;  // Copy probe attributes to info
    return err;
}

////////////////////////////////////////////////////////////
// Convert error code to descriptive string.

APIFUNC char* CCONV HVP_ErrorString(int errcode)
{
    switch (errcode)
    {
    case HVP_ERR_OK:            return "No errors";
    case HVP_ERR_DEVICE_MALLOC: return "Insufficient memory";
    case HVP_ERR_DEVICE_OPEN:   return "Error reading device info";
    case HVP_ERR_READ_OPEN:     return "Can't open device for reading";
    case HVP_ERR_WRITE_OPEN:    return "Can't open device for writing";
    case HVP_ERR_READ:          return "Error reading device";
    case HVP_ERR_WRITE:         return "Error writing device";
    case HVP_ERR_PROBE_ID:      return "Invalid probe number";
    case HVP_ERR_CHECKSUM:      return "Invalid calibration and S/N";
    case HVP_ERR_PROBE_CLOSED:  return "Probe is not open";
    case HVP_ERR_DEVICE_PARAM:  return "Unexpected device parameter";
    default:                    return "Unknown error code";
    }
}

////////////////////////////////////////////////////////
// Read API version numbers.

APIFUNC int CCONV HVP_GetApiVersion(void)
{
    return (VERSION_MAJOR << 24) | (VERSION_MINOR << 16) | VERSION_BUILD;
}


#ifdef __cplusplus
}
#endif

